// buha.cc
// version 2.5, SM & FTK, January 19, 2022
//              history: version 1.0, Lg, July 20, 1995
//                       version 1.1, Lg, August 2, 1996
//                       version 1.2, Lg, Nov. 21, 1996
//                       version 1.3, FTK, 15 May 2001
//                       version 1.5, Lg, Feb. 20, 2005
//                       version 1.6, PS, Sept 2011
//                       version 2.0, Lg, July 9, 2017
//                       version 2.1, SM, July 10, 2017
//                       version 2.3, SM, December 3, 2017
//                       version 2.4, SM, June 22, 2021

#include <iostream>
#include <fstream>
#include <string>
#include "uti1a.h"
#include "lineInput.h"
#include "menu.h"
#include "buha.h"
using namespace std;


const char* mid = "buha (January 2022, v2.5)";
const char* prompt = "buha> ";
const char* defaultfilename = "default";

long int line_count=-1;
menu men( &line_count );

istream* inp = &cin;
ostream* outp = &cout;

int fileinput=0;     // boolean
int _terminate=0;     // boolean

static void quit() { _terminate=1; }

static void skipline() { li_skipline(*inp); }

static void syntax_error()
{
    input_error( line_count, "syntax error" );
    skipline();
}

//////////////////////////////////////////////////////////////////////////

journal journ;
account_list accounts;

//////////////////////////////////////////////////////////////////////////

static int is_legal_in_filename( char ch )
{
    return isalpha(ch) || isdigit(ch)
           || (ch=='/')
           || (ch=='_')
           || (ch=='.')
           || (ch==':')
           || (ch=='~');
}


static int is_legal_in_account_name( char ch )
{
    return isalpha(ch) || isdigit(ch) 
           || (ch=='/')
           || (ch=='_')
           || (ch=='.');
}


static int is_legal_in_comment( char ch )
{
    return ch!='#';
}

///////////////////////////////////////////////////////////////////////////

class infile_access 
{ //  -------------
  private:
    ifstream *inp;
    long int line_count;
    infile_access *prev;
  friend void infile();
  friend void defaultinfile();
  friend void closefile();
};

int nesting_depth=0;
infile_access *fap=0;
ifstream *cinp=0;   // current file

void infile()
{
    const int max_files = 10;    // change freely, just here to 
                                 // prevent accidental recursion
    if ( nesting_depth >= max_files ) {
        input_error( line_count, "nested file input beyond maximum depth" );
        skipline(); return;
    }
    li_eatspace(*inp);
    enum { bufsize=256 };
    char filename[bufsize];
    int nchar = 
    li_get_string(*inp, filename, bufsize, &is_legal_in_filename, (int)line_count); 
    if ( nchar>=bufsize ) { skipline(); return; }
    if ( nchar==0 ) { input_error( line_count, "no filename" ); return; }
    ifstream *_inp = new ifstream(filename);
    if ( _inp==0 ) memory_error();  // might depend on C++ implementation
    if ( !_inp->is_open() ) {
        input_error( line_count, "can't open ", filename );
        delete _inp;
        return;
    }
    infile_access *_fap = fap;
    fap = new infile_access; 
    if ( fap==0 ) memory_error();
    //
    //*outp << "opening file " << filename 
    //      << " at level " << nesting_depth << endl;
    //
    fap->inp = cinp;
    fap->line_count = line_count;
    fap->prev = _fap;
    inp = cinp = _inp;
    line_count=1;
    fileinput=1;
    nesting_depth++;
}


void defaultinfile()
{
    const int max_files = 10;    // change freely, just here to 
                                 // prevent accidental recursion
    if ( nesting_depth >= max_files ) {
        input_error( line_count, "nested file input beyond maximum depth" );
        skipline(); return;
    }
    li_eatspace(*inp);
    ifstream *_inp = new ifstream(defaultfilename);
    if ( _inp==0 ) memory_error();  // might depend on C++ implementation
    if ( !_inp->is_open() ) {
        input_error( line_count, "can't open ", defaultfilename );
        delete _inp;
        return;
    }
    infile_access *_fap = fap;
    fap = new infile_access; 
    if ( fap==0 ) memory_error();
    fap->inp = cinp;
    fap->line_count = line_count;
    fap->prev = _fap;
    inp = cinp = _inp;
    line_count=1;
    fileinput=1;
    nesting_depth++;
}


void closefile()
{
    if ( fap==0 ) return;    // no file open
    delete cinp;
    cinp = fap->inp;
    inp = (cinp==0) ? &cin : cinp;
    line_count = fap->line_count;
    infile_access *prev = fap->prev;
    delete fap;
    fap = prev;
    nesting_depth--;
    if ( nesting_depth==0 ) fileinput=0;
    //
    //*outp << "file closed, continueing at level " << nesting_depth;
    //if ( nesting_depth>0 ) *outp << " at line " << line_count;
    //*outp << endl;
}

////////////////////////////////////////////////////////////////////////////

static void show_date()
{
    *outp << journ.day << " / " << journ.month << " / " << journ.year
          << "  (day / month / year)\n";
}


static void get_date()
{
    unsigned int day, month, year;
    if ( !li_get_unumber(*inp, day) ) {
        show_date();
        return;
    }
    if ( day>31 ) {
        input_error( line_count, "day > 31" );
        skipline(); return;
    }
    journ.day=day;
    li_eatspace(*inp); 
    char ch;
    if ( !li_get(*inp, ch) ) return;
    if ( ch!='/' ) { inp->putback(ch); return; }
    if ( !li_get_unumber(*inp, month) ) { 
        input_error( line_count, "month expected" ); 
        skipline(); return;
    }
    if ( month>12 ) {
        input_error( line_count, "month > 12" );
        skipline(); return;
    }
    journ.month=month;
    li_eatspace(*inp);
    if ( !li_get(*inp,ch) ) return; 
    if ( ch!='/' ) { inp->putback(ch); return; }
    if ( !li_get_unumber(*inp, year) ) { 
        input_error( line_count, "year expected" ); 
        skipline(); return;
    }
    journ.year=year;
}


static void date()
{
    int _day = journ.day; 
    int _month = journ.month;
    int _year = journ.year;
    int _withinday = journ.withinday;
    get_date();
    journ.withinday = 0;
    for (;;) {
        if ( journ.year > _year ) return; 
        if ( journ.year < _year ) break;
        if ( journ.month > _month ) return; 
        if ( journ.month < _month ) break;
        if ( journ.day > _day ) return;
        if ( journ.day < _day ) break;
        journ.withinday = _withinday; return; 
    }
    input_warning( line_count, "date goes backwards" );
}

////////////////////////////////////////////////////////////////////////

static account* get_account( int no_account_is_error=0 )
{
    const int bufsize = 80;
    char buf[bufsize];
    li_eatspace(*inp);
    int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
    if ( nchars >= bufsize ) { skipline(); return 0; }
    if ( nchars==0 ) {
        if ( no_account_is_error ) {
            input_error( line_count, "account expected" ); 
            skipline();
        }
        return 0;
    }
    int status;
    account* ap = (account*) accounts.retrieve( buf, &status );
    if ( status != simpleStringList::exact_match ) { 
        input_error( line_count, "unknown account ", buf );
        skipline(); return 0; 
    }
    return ap;
}

//////////////////////////////////////////////////////////////////////////

static void booking()
{
    account* lap = get_account();
    if ( lap==0 ) return;
    li_eatspace(*inp);
    char ch;
    if ( !li_get(*inp,ch) || (ch!='-') ) { 
        input_error( line_count, "syntax error: missing \"-\"" ); 
        skipline(); return;
    }
    account* rap = get_account(1);
    if ( rap==0 ) return;
    if ( lap->dec() != rap->dec() ) {
        input_error( line_count, "incompatible accounts" );
        skipline(); return;
    }
    buha_number amount; 
    if ( !li_get_unumber( *inp, amount ) ) {
        input_error( line_count, "missing amount" );
        skipline(); return;
    }
    amount *= lap->dec_pow();
    if ( !li_get(*inp,ch) ) {
        journ.put_entry( lap, rap, amount );  // no comment
        return;
    }
    if ( ch=='.' ) {
        unsigned int cents=0;
        int ndigits = li_get_unumber( *inp, cents );
        if ( ndigits!=lap->dec() ) {
            input_error( line_count, 
                         "wrong number of decimal digits after \".\"" );
            skipline(); return;
        }
        amount = amount + cents;
        if ( !li_get(*inp,ch) ) {
            journ.put_entry( lap, rap, amount );  // no comment
            return;
        }
    }
    if ( (ch!=' ') && (ch!='\t') ) {
        syntax_error(); skipline(); return;
    }
    li_eatspace(*inp);
    //
    const int bufsize = 256;
    char buf[bufsize];
    li_get_string( *inp, buf, bufsize, &is_legal_in_comment, (int)line_count, 1 );
  //if ( strlen(buf) > journal::max_comment ) 
  //   input_warning( line_count, "too much comment" );
    li_get(*inp,ch);   // skip '#'
    char buf2[bufsize];
    li_get_string( *inp, buf2, bufsize, &is_legal_in_comment, (int)line_count, 1 );
  //if ( strlen(buf2) > journal::max_comment2 ) 
  //    input_warning( line_count, "too much comment after \"#\"" );
    if ( li_get(*inp,ch) ) {
        input_error( line_count, "too many \"#\"" ); 
        skipline(); return;
    }
    journ.put_entry( lap, rap, amount, buf, buf2 ); 
}

//////////////////////////////////////////////////////////////////////

static void saldo()
{
    account* from = get_account(1); 
    if (from==0) { skipline(); return; }
    li_eatspace(*inp);
    char ch;
    if ( !li_get(*inp,ch) || (ch!='>') ) { 
        input_error( line_count, "syntax error: missing \">\"" ); 
        skipline(); return;
    }
    account* to = get_account(1);
    if (to==0) { skipline(); return; }
    if ( from==to ) { 
        input_error( line_count, "accounts must be different" );
        skipline(); return;
    }
    li_eatspace(*inp);
    const int bufsize = 256;
    char buf[bufsize];
    li_get_string( *inp, buf, bufsize, &is_legal_in_comment, (int)line_count, 1 );
  //if ( strlen(buf) > journal::max_comment ) 
  //   input_warning( line_count, "too much comment" );
    buha_number diff = from->left_sum() - from->right_sum();
    if ( diff>=0 ) journ.put_entry( to, from, diff, buf );
    else journ.put_entry( from, to, -diff, buf );
}

static void rsaldo()
{
    account* from = get_account(1); 
    if (from==0) { skipline(); return; }
    li_eatspace(*inp);
    char ch;
    if ( !li_get(*inp,ch) || (ch!='>') ) { 
        input_error( line_count, "syntax error: missing \">\"" ); 
        skipline(); return;
    }
    account* to = get_account(1);
    if (to==0) { skipline(); return; }
    if ( from==to ) { 
        input_error( line_count, "accounts must be different" );
        skipline(); return;
    }
    li_eatspace(*inp);
    const int bufsize = 256;
    char buf[bufsize];
    li_get_string( *inp, buf, bufsize, &is_legal_in_comment, (int)line_count, 1 );
  //if ( strlen(buf) > journal::max_comment ) 
  //   input_warning( line_count, "too much comment" );
    buha_number diff = from->left_sum() - from->right_sum();
    if ( diff>0 ) journ.put_entry( to, from, diff, buf );
    else journ.put_entry( from, to, -diff, buf );
}

//////////////////////////////////////////////////////////////////////

static void define_account()
{
    const int bufsize=256;
    char buf[bufsize];
    li_eatspace(*inp);
    int nchars = 
    li_get_string( *inp, buf, bufsize, &is_legal_in_account_name, (int)line_count );
    if ( nchars >= bufsize ) { skipline(); return; }
    if ( nchars==0 ) {
        input_error( line_count, "account name expected" );
        skipline(); return; 
    }
    int status;
    simpleStringListEntry* slep;
    slep = accounts.retrieve( buf, &status );
    if ( status == simpleStringList::exact_match ) {
        input_error( line_count, "redefinition of account ", buf );
        return;
    }
    char* name = new char[strlen(buf)+1];
    if ( name==0 ) memory_error();
    strcpy( name, buf );
    li_eatspace(*inp);
    li_get_string( *inp, buf, bufsize, &is_legal_in_comment, (int)line_count, 1 );
    account *accp = new account(name, buf);
    if ( accp==0 ) memory_error();
    accounts.insert( accp, slep );
}

//////////////////////////////////////////////////////////////////////

static void reset()
{
    journ.clear();
    accounts.clear(1);
}

////////////////////////////////////////////////////////////////////

static void show() 
{ 
    account* ap = get_account();
    if (ap) ap->print(*outp);
    else accounts.print_account_list(*outp);
}


static void Export() 
{ 
    account* ap = get_account();
    if (ap) ap->Export(*outp);
    else accounts.export_account_list(*outp);
}


static void line() 
{ 
  *outp << "\n";
}


static void stop() 
{ 
  //*outp << "Gugus";
  closefile();
}


static void LaTeX()
{
    account* ap = get_account();
    if (ap) ap->print_LaTeX(*outp);
}


static void _printAll(int format=0)
{
    li_eatspace(*inp);
    enum { bufsize=256 };
    char foldername[bufsize];
    int nchar = 
    li_get_string(*inp, foldername, bufsize, &is_legal_in_filename, (int)line_count); 
    if ( nchar>=bufsize ) { skipline(); return; }
    if ( nchar==0 ) { input_error( line_count, "no folder" ); return; }
    *outp << "writing files to (presumably) existing folder " << foldername << "..." << endl;
	accounts.print_accounts_tofiles(foldername, format);
}


static void printAll()
{
    _printAll();
}


static void printAllCSV()
{
    _printAll(1);
}


static void printAllLaTeX()
{
    _printAll(2);
}


static void help() 
{
    *outp << endl << mid << endl 
          << "available commands:\n\n"; 
    men.print_command_list(*outp, 1, 19); 
    *outp << endl;
    *outp << "see also \\show\n"; 
}


static void init_menu()
{
    men.def( "\\", &syntax_error, "", 1 );
    men.def( "\\help", &help, "print this list of commands" );
    //
    men.def( "\\date", &date,
	     "show or set date (example: \\date 13/7/2017)" );
    men.def( "\\new", &define_account, 
             "define new account (example: \\new Deb)" );
    men.def( "\\saldo", &saldo,
	     "balance an account (example: \\saldo Ka > Bilanz)" );
    men.def( "\\rsaldo", &rsaldo,
	     "like \\saldo, but a zero is booked on the right instead of left" );
    men.def( "\\infile", &infile,
	     "read file (example: \\infile myfile)" );
    men.def( "\\\\", &defaultinfile,
	     "read file called 'default' (for default commands)" );
    men.def( "\\show", &show,
	     "with account: show account (example: \\show Ka);\n\t\twithout account: show list of accounts" );
    men.def( "\\export", &Export,
	     "like \\show, but only totals are shown and no account names\n\t\tor comments are given (useful for copy-paste)" );
    //
    men.def( "\\printAll", &printAll,
             "print all accounts, each as a separate text file, into specified\n\t\tfolder (folder must exist) (example: \\printAll myfolder)");
    men.def( "\\printAllCSV", &printAllCSV,
             "print all accounts, each as a separate CSV file, into specified\n\t\tfolder (folder must exist) (example: \\printAllCSV myfolder)");
    men.def( "\\printAllLaTeX", &printAllLaTeX,
             "print all accounts, each as a separate LaTeX file, into specified\n\t\tfolder (folder must exist) (example: \\printAllLaTeX myfolder)");
    men.def( "\\line", &line, "print empty line" );
    //
    men.def( "\\stop", &stop, "ignore rest of file" );
    men.def( "\\reset", &reset, "erase everything" );
    men.def( "\\quit", &quit, "exit program" );
    //
}

//////////////////////////////////////////////////////////////////////

int main()
{
    init_menu();
    char ch;
    *outp << mid << " (type \"\\help\" for help)\n" << prompt;
    while ( *inp && *outp && !_terminate ) {
        li_eatspace(*inp);
        if ( inp->get(ch)==0 ) { 
            closefile();
            continue;
        }
        if ( (ch=='\n') || (ch=='\r') ) {
            if (fileinput) line_count++;
            else *outp << prompt;
            continue;
        }
        inp->putback(ch);
        if ( ch=='\\' ) men.get_and_execute_command(*inp);
        else if ( is_legal_in_account_name(ch) ) booking();
        else {
            if ( ch!='%' ) input_error( line_count, "illegal character" );
            skipline();
        }
    }
    return 0;
}
