// buha.h
// version 2.1, SM, July 10, 2017
//              history: version 1.0, Lg, July 19, 1995
//                       version 1.1, Lg, July 30, 1996
//                       version 1.2, Lg, Nov. 21, 1996
//                       version 1.3, FTK, May 15, 2001
//                       version 1.6, PS, Sept 2011
//                       version 2.0, Lg, July 9, 2017



#ifndef BUHA_H
#define BUHA_H

#include "simpleList.h"


typedef long int buha_number;      // all in cents
//               ===========       // position of decimal point
                                   // depending on account

class buha_omanip; // forward
class account;     // forward


class journal_entry: public simpleListEntry {
//    =============
  friend class journal;
  friend class account;
  protected:
    int _year, _month, _day, _withinday; 
    account* left_account;
    account* right_account;
    buha_number amount; 
    char* _comment;
    char* _comment2; // proof
};


class journal: public simpleList {
//    =======
  public:
    int year, month, day, withinday; 
    journal() { year=month=day=withinday=0; }
    void put_entry( account* left, account* right, buha_number amnt,
                    char* comment=0, char* comment2=0 ); // comments are copied 
    enum { max_comment=24, max_comment2=7 };  // comments may be larger, but 
                                              // listings might look ugly
    void clear();    // delete whole list
    ~journal() { clear(); }
};


class account_entry: public simpleListEntry {
//    =============
  public:
    journal_entry* jep;
};


class account: public simpleList, public simpleStringListEntry {
//    =======
  friend class journal;
  protected:
    buha_number _left_sum, _right_sum;
    int _dec, _dec_pow; 
    char* _comment;
    ostream& print_buha_number_LaTeX( ostream&, buha_number );
  public:
    account( char* name, char* comment=0, int decimals=2 ); //comment is copied
    buha_number left_sum() { return _left_sum; }
    buha_number right_sum() { return _right_sum; }
    int dec() { return _dec; }
    int dec_pow() { return _dec_pow; }
    buha_omanip buhaprint( buha_number );
    ostream& print_buha_number( ostream&, buha_number );
    void print( ostream&, int extraspace=0 );
    void Export( ostream& );
    void print_csv( ostream&);  // for csv files
    void print_LaTeX( ostream&, int mode=0 );  // mode=1: left-right sorted
    void clear();   // delete whole list 
    ~account();     // delete whole list and comment, but not account name
};


class account_list: public simpleStringList {
//    ============
  public:
    account_list() {;}
    void print_account_list( ostream& os, int tab=15, int width=10 ); // print only names and totals
    void export_account_list( ostream& os, int tab=15, int width=10 ); // print only totals
    void print_accounts_tofiles( const char* foldername, int format=0 );  
        // print full accounts, to text files; format: 0 for txt, 1 for csv, 2 for LaTeX
    void clear( int delete_account_names=0 );     // delete whole list
    ~account_list() { clear(); }
};


class buha_omanip {
//    ===========
  private:
    account* _ap;
    buha_number _x;
  public:
    buha_omanip(account* ap, buha_number x ): _ap(ap), _x(x) {};
  friend
    ostream& operator<<(ostream& os, buha_omanip m)
        { return (m._ap)->print_buha_number(os, m._x); }
};


#endif
