// buhaAccounts.cc
// version 2.5, SM & FTK, January 19, 2022
//              history: version 1.0, Lg, July 19, 1995
//                       version 1.1, Lg, July 30, 1996
//                       version 1.2, Lg, Nov. 21, 1996
//                       version 1.6, PS, Sept 2011
//                       version 2.0, Lg, July 9, 2017
//                       version 2.1, SM, July 10, 2017

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string.h>
#include "uti1a.h"
#include "buha.h"
using namespace std;

static char* empty_string = (char*)"";


void journal::put_entry( account* left, account* right, buha_number amnt, 
/*            ========= */                     char* comment, char* comment2 )
{
    journal_entry* p = new journal_entry;
    if ( p==0 ) memory_error();
    append(p);
    p->_year = year; p->_month = month; 
    p->_day = day; p->_withinday = ++withinday;
    p->left_account = left; 
    p->right_account = right;
    p->amount = amnt;
    if (comment==0) p->_comment = empty_string;
    else {
        p->_comment = new char[strlen(comment)+1];
        if ( p->_comment==0 ) memory_error();
        strcpy( p->_comment, comment );
    }
    if (comment2==0) p->_comment2 = empty_string;
    else {
        p->_comment2 = new char[strlen(comment2)+1];
        if ( p->_comment2==0 ) memory_error();
        strcpy( p->_comment2, comment2 );
    }
    account_entry* laep = new account_entry;
    if ( laep==0 ) memory_error();
    laep->jep = p;
    left->append(laep);
    if ( left!=right ) { 
        account_entry* raep; 
        raep = new account_entry; 
        if ( raep==0 ) memory_error();
        raep->jep = p;
        right->append(raep); 
    }
    left->_left_sum += amnt; right->_right_sum += amnt;
}


void journal::clear()
{ //          =====
    journal_entry *p = (journal_entry*)_first;
    while (p) {
        journal_entry *next = (journal_entry*)p->_next;
        if (p->_comment!=empty_string) delete [] p->_comment;
        if (p->_comment2!=empty_string) delete [] p->_comment2;
        delete p;
        p = next;
    }
    _list_length = 0;
    _first = _last = 0;
    year=month=day=withinday=0;
}


account::account( char* name, char* comment, int decimals )
{ //     =======
    _left_sum = _right_sum = 0; 
    strngp = name; 
    _dec = decimals;
    _dec_pow = power( 10, decimals );
    if ( comment==0 ) _comment = empty_string; 
    else {
        _comment = new char[strlen(comment)+1];
        if ( _comment==0 ) memory_error();
        strcpy( _comment, comment );
    }
}


void account::clear()
{ //          =====
    account_entry *p = (account_entry*)_first;
    while (p) {
        account_entry *next = (account_entry*)p->_next;
        delete p;
        p = next;
    }
    _list_length = 0;
    _first = _last = 0;
}


account::~account()
{ //     ========
    clear();
    if ( _comment!=empty_string ) delete[] _comment;
}


void account_list::clear( int delete_account_names )
{ //               =====
    account *p = (account*)_first;
    while (p) {
        account *next = (account*)p->_next;
        if (delete_account_names) delete[] (char*)p->strngp;
        delete p;
        p = next;
    }
    _list_length = 0;
    _first = _last = 0;
}


ostream& account::print_buha_number( ostream& os, buha_number x )
{ //              =================
    int minus = (x<0);
    if ( minus ) x = -x;
    buha_number y = x / _dec_pow;
    if (minus) {
        if (y) os << -y;
        else os << "-0";
    }
    else os << y; 
    int cents = (int)x % _dec_pow;
    if ( cents )
        os << "." << setw(_dec) << setfill('0') 
           << cents << setfill(' ');
    else os << setw(_dec+1) << "";
    return os;
}


buha_omanip account::buhaprint( buha_number x )
{ //                 =========
    return buha_omanip(this, x);
}


ostream& account::print_buha_number_LaTeX( ostream& os, buha_number x )
{ //              =======================
    os << "{";
    int minus = (x<0);
    if ( minus ) x = -x;
    buha_number y = x / _dec_pow;
    if (minus) {
        if (y) os << -y;
        else os << "-0";
    }
    else os << y; 
    os << "}{";
    int cents = (int)x % _dec_pow;
    if ( cents )
        os << "." << setw(_dec) << setfill('0') 
           << cents << setfill(' ');
    os << "}";
    return os;
}


void account::print( ostream& os, int extraspace)
{ //          =====
    const char *margin = "";
    if (extraspace & 1) { 
        // margin = "    ";   // as you please
        os << endl << endl;
    }
    os << margin << strngp << ": " << _comment << endl;
    if (extraspace & 1) { os << endl; }
    simpleListEntry* p = _first;
    account_entry* aep;
    journal_entry* jep;
    account* otherp;
    while (p) {
        aep = (account_entry*) p;
        jep = aep->jep;
        os << margin
           << adjust_right
           << setw(2) << jep->_day << '/'
           << setw(2) << jep->_month << '/'
           << setw(4) << jep->_year << " ." 
           << adjust_left 
           << setw(4) << jep->_withinday; 
        if ( (otherp = jep->left_account) == this )
            otherp = jep->right_account;
        os << setw(10) << otherp->strngp 
           << adjust_right
           << setw(8); 
        if ( jep->left_account == this ) os << buhaprint(jep->amount);
        else os << "" << setw(_dec+1) << "";
        os << setw(10);
        if ( jep->right_account == this ) os << buhaprint(jep->amount);
        else os << "" << setw(_dec+1) << "";
	int outDiff = (int)strlen(jep->_comment);
	for (const char* text=jep->_comment; *text!='\0'; ++text) {
	  outDiff -= (*text&0xC0) != 0xC0;
	}
	os << "   " 
           << adjust_left 
           << setw(journal::max_comment+outDiff) << jep->_comment;
        if ( *(jep->_comment2) ) {
            os << " #" << jep->_comment2; 
        }
        os << adjust_right << endl;
        p = p->_next;
    }
    os << margin << setw(16) << "";
    os << "total:    "
       << setw(8) << buhaprint(_left_sum) << " -" 
       << setw(8) << buhaprint(_right_sum) << " ="
       << setw(8) << buhaprint(_left_sum - _right_sum) 
       << endl;
}


void account::Export( ostream& os )
{ //          ======
    os << strngp << ": " << _comment << endl;
    simpleListEntry* p = _first;
    account_entry* aep;
    journal_entry* jep;
    account* otherp;
    while (p) {
        aep = (account_entry*) p;
        jep = aep->jep;
        //os << jep->_day << '/'
        //   << jep->_month << '/'
        //   << jep->_year << "\t." 
        //   << jep->_withinday; 
        //if ( (otherp = jep->left_account) == this )
        //    otherp = jep->right_account;
        //os << "\t" << otherp->strngp 
        //   << "\t";
        //if ( jep->left_account == this ) os << buhaprint(jep->amount);
        //else os << "\t";
        //os << "\t";
        //if ( jep->right_account == this ) os << buhaprint(jep->amount);
        //else os << "\t";
        //os << "\t" << jep->_comment;
        //if ( *(jep->_comment2) ) {
        //    os << "\t#" << jep->_comment2; 
        //}
	os << buhaprint(jep->amount);
        os << endl;
        p = p->_next;
    }
    //os << "total:\t"
    //   << buhaprint(_left_sum) << " - " 
    //   << buhaprint(_right_sum) << " = "
    //   << buhaprint(_left_sum - _right_sum) 
    //   << endl;
    os << endl << buhaprint(_left_sum) << endl;

}


void account::print_csv( ostream& os)
{ //          =========
    os << endl << endl << strngp << ": " << _comment << endl << endl;
    simpleListEntry* p = _first;
    account_entry* aep;
    journal_entry* jep;
    account* otherp;
    while (p) {
        aep = (account_entry*) p;
        jep = aep->jep;
        os << jep->_day << '/'
           << jep->_month << '/'
           << jep->_year << " ." 
           << jep->_withinday << ",   "; 
        if ( (otherp = jep->left_account) == this )
            otherp = jep->right_account;
        os << otherp->strngp << ",  ";
        if ( jep->left_account == this ) os << buhaprint(jep->amount);
        else os << "" << setw(_dec+1) << "";
        os << ",  ";
        if ( jep->right_account == this ) os << buhaprint(jep->amount);
        else os << "" << setw(_dec+1) << "";
        os << ",   , " << jep->_comment;
        if ( *(jep->_comment2) ) {
            os << ", #" << jep->_comment2; 
        }
        os << endl;
        p = p->_next;
    }
    os << "         , total:,  "
       << buhaprint(_left_sum) << ",  " 
       << buhaprint(_right_sum) << ",  "
       << buhaprint(_left_sum - _right_sum) 
       << endl;
}


void account::print_LaTeX( ostream& os, int mode )  // mode=0: unsorted
{ //          ===========                        // mode=1: left-right sorted
                                                 // mode=2: right only
    if ( mode<2 ) {
        if ( mode==0 ) os << "\\begin{account}{";
        else os << "\\begin{LRaccount}{";
        os << strngp << "}{" << _comment << "}\n";
    }
    simpleListEntry* p = _first;
    account_entry* aep;
    journal_entry* jep;
    account* otherp;
    for ( ; p; p=p->_next ) {
        aep = (account_entry*) p;
        jep = aep->jep;
        if ( jep->left_account == this ) {
            if ( mode==2 ) continue;
            os << "\\Litem{";
            otherp = jep->right_account;
        }
        else {
            if ( mode==1 ) continue;
            os << "\\Ritem{";
            otherp = jep->left_account;
        }
        os << jep->_day << "}{" << jep->_month 
           << "}{" << jep->_year << "}{" << jep->_withinday << "}{"
           << otherp->strngp << "}";
        print_buha_number_LaTeX(os, jep->amount);
        os << "{" << jep->_comment << "}{" << jep->_comment2 << "}\n";
    }
    if ( mode<2 ) {
        os << "\\Ltotal";
        print_buha_number_LaTeX(os, _left_sum);
        os << endl;
    }
    if ( mode!=1 ) {
        os << "\\Rtotal";
        print_buha_number_LaTeX(os, _right_sum);
        os << endl;
    }
    if ( mode==1 ) print_LaTeX(os, 2);
    if ( mode<2 ) {
        if ( mode==0 ) os << "\\end{account}\n";
        else os << "\\end{LRaccount}\n";
    }
}


void account_list::print_account_list( ostream& os, int tab, int width )
{ //               ==================
    simpleListEntry* p = _first;
    account* ap;
    int space;
    buha_number left, right;
    os << adjust_right; 
    while (p) {
        ap = (account*) p;
        os << ap->strngp;
        space = (tab-3) - (int)strlen( ap->strngp );
        while ( space-- > 0 ) os << ' ';
        os << setw(width) << ap->buhaprint(left = ap->left_sum()) << "  - " 
           << setw(width) << ap->buhaprint(right = ap->right_sum()) << "  = ";
        os << setw(width) << ap->buhaprint(left - right) << endl;
        p = p->_next;
    }
}


void account_list::print_accounts_tofiles( const char* folder, int format )
{ //               ======================
    string filenamesuffix = ".txt";
    if (format==1) filenamesuffix = ".csv";
    else if (format==2) filenamesuffix = ".tex";
    simpleListEntry* p = _first;
    account* ap;
    while (p) {
        ap = (account*) p;
        ofstream file(string(folder) + "/" + ap->strngp + filenamesuffix);
		if (file.is_open()) {
		    if (format==1) ap->print_csv(file);
			else if (format==2) ap->print_LaTeX(file);
			else ap->print(file, 1);
			file.close();
		}
		else { 
			cerr << "failure" << endl;
			return;
		}
        p = p->_next;
    }
}


void account_list::export_account_list( ostream& os, int tab, int width )
{ //               ===================
    simpleListEntry* p = _first;
    account* ap;
    int space;
    buha_number left, right;
    os << adjust_right; 
    while (p) {
        ap = (account*) p;
        //os << ap->strngp << '\t';
	left = ap->left_sum();
	right = ap->right_sum();
	if ( ap->strngp[0] == 'E' ) { // if account name starts with E change sign
	  os << ap->buhaprint(right - left) << endl;
	} else {
	  os << ap->buhaprint(left - right) << endl;
	};
	p = p->_next;
    }
}
